package server;
import java.io.IOException;
import java.io.InterruptedIOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.ProtocolException;
import java.net.Socket;
import java.util.ArrayList;
import java.util.Random;

import org.xbill.DNS.DClass;
import org.xbill.DNS.Flags;
import org.xbill.DNS.Header;
import org.xbill.DNS.Message;
import org.xbill.DNS.OPTRecord;
import org.xbill.DNS.Opcode;
import org.xbill.DNS.Rcode;
import org.xbill.DNS.Record;
import org.xbill.DNS.Section;
import org.xbill.DNS.TXTRecord;
import org.xbill.DNS.Type;

import common.ClientPacket;
import common.Common;
import common.ConnectionAcceptedServerPacket;
import common.ConnectionClosedClientPacket;
import common.ConnectionClosedServerPacket;
import common.ConnectionRequestClientPacket;
import common.ConnectionState;
import common.LZMAPacketConverter;
import common.Log;
import common.PortCommunicator;
import common.ServerPacket;
import common.SpecialClientPacket;
import common.TCPPortCommunicator;
import common.UDPPortCommunicator;

public class Server {
	public static final String SERVER_NAME = "tcp-over-dns-server";
	public static final String DEFAULT_ADDR = "0.0.0.0";

	private ArrayList<ConnectionState> clients = new ArrayList<ConnectionState>();
	private InetSocketAddress forwardAddress;
	private ServerOptions serverOpts;
	private Random rand;
	
	public static void main(String[] args) {
		try {
			ServerCmdLineParser clp = new ServerCmdLineParser();
			ServerOptions opts = null;
			try {
				opts = clp.parseCmdLine(args);
			} catch (Exception e) {
				e.printStackTrace();
			}
			
			if (opts == null) {
				clp.printUsage();
			} else {
				Log.set(new Log(System.out, Log.integerToLevels(opts.logLevel)));
				new Server(opts);
			}
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	public static void logOptions(ServerOptions opts) {
		Log.get().println(Log.LEVEL_INFO, SERVER_NAME + " starting up");
		Log.get().println(Log.LEVEL_INFO, "Hosting domain: " + opts.domain);
		for (InetAddress dns : opts.listenAddrs) {
			for (Integer port : opts.listenPorts) {
				Log.get().println(Log.LEVEL_INFO, "DNS listening on: " + dns + ":" + port);
			}
		}
		Log.get().println(Log.LEVEL_INFO, "Forwarding to: " + opts.forwardAddr + ":" + opts.forwardPort);
		Log.get().println(Log.LEVEL_INFO, "MTU: " + opts.mtu);
		Log.get().println(Log.LEVEL_INFO, "Log level: " + opts.logLevel);
	}
	
	public Server(ServerOptions opts) throws IOException {
		logOptions(opts);
		
		serverOpts = opts;
		rand = new Random();
		
		forwardAddress = new InetSocketAddress(opts.forwardAddr, opts.forwardPort);
		
		for (InetAddress addr : opts.listenAddrs) {
			for (int port : opts.listenPorts) {
				addUDP(addr, port);
				//System.out.println(SERVER_NAME + ": listening on "
				//		+ addrport(addr, port));
			}
		}
		
		ClientTimeoutThread timeoutThread = new ClientTimeoutThread();
		timeoutThread.start();
		
		//System.out.println(SERVER_NAME + ": running");
	}

	/*
	 * Note: a null return value means that the caller doesn't need to do
	 * anything.
	 */
	byte[] generateReply(Message query, byte[] in, int length, Socket s)
			throws IOException {
		
		//pretend 70% packet loss
		//if (new Random().nextFloat() > .3) {
		//	return null;
		//}
		
		Header header = query.getHeader();
		
		//if they are sending us a query response then the response packets 
		//are going in the wrong direction
		if (header.getFlag(Flags.QR))
			return null;
		//if there was already an error, just report it
		if (header.getRcode() != Rcode.NOERROR)
			return errorMessage(query, Rcode.FORMERR);
		//we can only handle queries at this point
		if (header.getOpcode() != Opcode.QUERY)
			return errorMessage(query, Rcode.NOTIMP);

		//every response has the question in it
		Record queryRecord = query.getQuestion();

		int type = queryRecord.getType();
		//general case
		if (!Type.isRR(type) && type != Type.ANY) {
			return errorMessage(query, Rcode.NOTIMP);
		}

		//right now we only support A
		if (type != Type.TXT) {
			return errorMessage(query, Rcode.NOTIMP);
		}

		Message response = new Message(query.getHeader().getID());
		response.getHeader().setFlag(Flags.QR);
		response.getHeader().setFlag(Flags.RD);
		response.getHeader().setFlag(Flags.RA);
		response.getHeader().setFlag(Flags.AA);
		response.addRecord(queryRecord, Section.QUESTION);
		
		ClientPacket clientPkt;
		ServerPacket responsePkt;
		try {
			
			clientPkt = ClientPacket.decodeClientPacket(queryRecord.getName(), serverOpts.domain);
			Log.get().println(Log.LEVEL_SPAM, "Received: " + clientPkt);

			if (!clientPkt.isChecksumValid()) {
				throw new ProtocolException("Bad Checksum");
			}
			
			//special protocol message
			if (clientPkt.isSpecial()) {
				try {
					responsePkt = handleSpecialPacket((SpecialClientPacket) clientPkt);
				} catch (Exception e) {
					Log.get().exception(Log.LEVEL_ERROR, e);
					return errorMessage(query, Rcode.FORMERR);
				}
			} else {//general data message
				ConnectionState client = getClient(clientPkt.getClientID());
				if (client == null) {
					System.out.println("Message from unknown client! ClientID: " + (0xff&clientPkt.getClientID()));
					responsePkt = new ConnectionClosedServerPacket(clientPkt.getClientID());
				} else {
					//its possible that 2 messages could come in on different UDP serving threads at the same time
					synchronized (client) {
						client.dnsPacketReceived(clientPkt);
						responsePkt = (ServerPacket) client.update();
					}
				}
			}
		} catch (Exception e) {
			Log.get().exception(Log.LEVEL_ERROR, e);
			return errorMessage(query, Rcode.FORMERR);
		}

		//byte [] randAddr = new byte[4];
		//rand.nextBytes(randAddr);
		responsePkt.updateChecksum();
		TXTRecord txtOut = responsePkt.encodeAsTxt(queryRecord.getName(), DClass.IN, Common.DEFAULT_TTL);
		assert (null != ServerPacket.decodeServerPacket(txtOut));
		response.addRecord(txtOut, Section.ANSWER);
		
		Log.get().println(Log.LEVEL_SPAM, "Sent: " + responsePkt);

		return response.toWire(determineMaxResponseLength(s, query.getOPT()));
	}

	private ServerPacket handleSpecialPacket(SpecialClientPacket clientPkt) throws Exception {
		switch (clientPkt.getType()) {
		case SpecialClientPacket.SPECIAL_TYPE_CONNECTION_REQUEST:
			ConnectionRequestClientPacket crcp = (ConnectionRequestClientPacket) clientPkt; 
			byte clientId = getOpenClientId();
			int mtu = Math.min(crcp.getMaxMTU(), serverOpts.mtu);
			boolean tcp = (crcp.getRequestFlags() & ConnectionRequestClientPacket.FLAG_TCP) != 0;
			
			PortCommunicator comm;
			if (tcp) {
				Socket tcpSock = new Socket(serverOpts.forwardAddr, serverOpts.forwardPort);
				tcpSock.setTcpNoDelay(true);
				tcpSock.setReceiveBufferSize(Common.DEFAULT_MTU);
				comm = new TCPPortCommunicator(tcpSock);
				Log.get().println(Log.LEVEL_INFO, "New tcp client connection:" + (clientId&0xff));
			} else {
				UDPPortCommunicator udpComm = new UDPPortCommunicator(new DatagramSocket(0), mtu, crcp.getLateTimer()*2, false); 
				udpComm.addPortForward(forwardAddress);
				comm = udpComm;
				Log.get().println(Log.LEVEL_INFO, "New udp client connection:" + (clientId&0xff));
			}
			
			ConnectionState newConnection = new ConnectionState(clientId, 
					new LZMAPacketConverter(),
					new ServerPacketFactory(serverOpts.domain),
					comm,
					crcp.getLateTimer());
			
			//start listening AFTER the listeners have been added by the connection state
			comm.start();
			
			synchronized (clients) {
				clients.add(newConnection);
			}
			ConnectionAcceptedServerPacket cap = new ConnectionAcceptedServerPacket(clientId, mtu, crcp.getChallengeId());
			return cap;
		case SpecialClientPacket.SPECIAL_TYPE_CONNECTION_CLOSED:
			ConnectionClosedClientPacket closeReq = (ConnectionClosedClientPacket) clientPkt; 
			ConnectionState client = removeClient(closeReq.getClientID());
			if (client != null) {
				client.close();				
			}
			//whether or not the client existed, respond with close
			ConnectionClosedServerPacket conClose = new ConnectionClosedServerPacket(closeReq.getClientID());
			return conClose;
		default:
		}
		throw new ProtocolException("Unknown special packet type");
	}

	//TODO make this a hash?
	private ConnectionState getClient(byte clientId) {
		synchronized (clients) {
			for (ConnectionState c : clients) {
				if (clientId == c.getClientId()) {
					return c;
				}
			}
		}
		return null;
	}
	
	private ConnectionState removeClient(byte clientID) {
		synchronized (clients) {
			for (int i=0; i<clients.size(); i++) {
				ConnectionState client = clients.get(i);
				if (client.getClientId() == clientID) {
					clients.remove(i);
					return client;
				}
			}
		}
		return null;
	}

	//TODO remove a synchronization somewhere
	private byte getOpenClientId() throws Exception {
		synchronized (clients) {
			int start = this.rand.nextInt(255);
			for (int i=0; i<256; i++) {
				if (null == getClient((byte)(start+i))) {
					return (byte)(start + i);
				}
			}
		}
		throw new Exception("No Open Client Slots!?");
	}
	
	private void timeoutClients() {
		synchronized (clients) {
			for (int i=clients.size()-1; i>=0; i--) {
				ConnectionState client = clients.get(i);
				if (client.hasTimedOut(Common.CONN_TIMEOUT)) {
					Log.get().println(Log.LEVEL_INFO, "Client timeout (ClientID:" + (0xFF&client.getClientId()) + ").");
					client.close();
					clients.remove(client);
				}
			}
		}
		
	}
	
	private static int determineMaxResponseLength(Socket s, OPTRecord queryOPT) {
		/*int maxLength;
		if (s != null) {
			maxLength = 65535;
		} else if (queryOPT != null) {
			maxLength = Math.max(queryOPT.getPayloadSize(), 512);
		} else {
			maxLength = 512;
		}
		return maxLength;
		*/
		return 512;
	}

	byte[] buildErrorMessage(Header header, int rcode, Record question) {
		Message response = new Message();
		response.setHeader(header);
		for (int i = 0; i < 4; i++)
			response.removeAllRecords(i);
		if (rcode == Rcode.SERVFAIL)
			response.addRecord(question, Section.QUESTION);
		header.setRcode(rcode);
		return response.toWire();
	}

	public byte[] formerrMessage(byte[] in) {
		Header header;
		try {
			header = new Header(in);
		} catch (IOException e) {
			return null;
		}
		return buildErrorMessage(header, Rcode.FORMERR, null);
	}

	public byte[] errorMessage(Message query, int rcode) {
		return buildErrorMessage(query.getHeader(), rcode, query.getQuestion());
	}

	private void serveUDP(InetAddress addr, int port) {
		DatagramPacket dpOut = new DatagramPacket(new byte[1], 0);
		byte[] in = new byte[512];
		DatagramPacket dpIn = new DatagramPacket(in, in.length);

		try {
			DatagramSocket sock = new DatagramSocket(port, addr);

			while (true) {
				dpIn.setLength(in.length);
				try {
					sock.receive(dpIn);
				} catch (InterruptedIOException e) {
					continue;
				}
				
				byte[] response = null;
				try {
					Message query = new Message(in);
					//System.out.println(query.toString());
					response = generateReply(query, in, dpIn.getLength(), null);
				} catch (IOException e) {
					Log.get().exception(Log.LEVEL_ERROR, e);
					response = formerrMessage(in);
				}

				//no response? dont send anything
				if (response == null) {
					continue;
				}
				
				dpOut.setData(response);
				dpOut.setLength(response.length);
				dpOut.setAddress(dpIn.getAddress());
				dpOut.setPort(dpIn.getPort());
				//System.out.println("sending to:" + dpOut.getAddress() + " port:" + dpOut.getPort());
				sock.send(dpOut);
			}
		} catch (IOException e) {
			Log.get().exception(Log.LEVEL_INFO, e);
		}
	}
	
	public void addUDP(final InetAddress addr, final int port) {
		Thread t = new Thread("DNS Serve " + addr + ":" + port) {
			public void run() {
				serveUDP(addr, port);
			}
		};
		t.start();
	}

	private class ClientTimeoutThread extends Thread {

		public ClientTimeoutThread() {
			super("Client timeout");
		}

		@Override
		public void run() {
			while (true) {
				try {
					sleep(1000);
				} catch (InterruptedException e) {
				}
				
				timeoutClients();
			}
		}

	}

}
